import importlib.metadata

from .univariate import univariate as univariate
from .univariate import univariate_importance as univariate_importance

from .multivariate import multivariate as multivariate

__version__ = importlib.metadata.version("ghq")

del importlib



from typing import Callable
from itertools import product
from functools import partial

from numpy.polynomial.hermite_e import hermegauss
from jax import numpy as jnp, jit, Array, vmap


@partial(jit, static_argnums=(0, 3))
def multivariate(
    integrand: Callable[[Array], Array],
    mean: Array,
    cov: Array,
    degree: int = 32,
) -> float:
    """
    Multivariate Gauss-Hermite integration.
    https://en.wikipedia.org/wiki/Gauss-Hermite_quadrature

    Not vectorised in mean and cov, but works with jax.vmap.

    .. math::

        \int integrand(x) \text{N}(x | mean, cov) dx

    Args:
        integrand: Function to be integrated over f(x).
        mean: Mean of multivariate Gaussian.
        cov: Covariance matrix of multivariate Gaussian.
        degree: Integer number of Gauss-Hermite points, defaults to 32.
    Returns:
        out: Approximate Gaussian expectation (scalar float).
    """
    cov_sqrt = jnp.linalg.cholesky(cov)
    dim = mean.size

    def standardized_integrand(x):
        return integrand(mean + cov_sqrt @ x)

    x, w = hermegauss(degree)

    xn = jnp.array(list(product(*(x,) * dim)))
    wn = jnp.prod(jnp.array(list(product(*(w,) * dim))), 1)

    hx = vmap(standardized_integrand)(xn)
    wn = wn.reshape((degree**dim,) + (1,) * (hx.ndim - 1))
    return (wn * hx).sum(0) / jnp.sqrt(2 * jnp.pi) ** dim



from jax import numpy as jnp, jit

import ghq


def test_zero():
    def f(x):
        return 0.0

    mean = jnp.array([0.0])
    cov = jnp.array([[1.0]])
    degree = 32
    result = jit(ghq.multivariate, static_argnums=(0, 3))(f, mean, cov, degree)
    assert jnp.isclose(result, 0.0)

    mean = jnp.array([0.0, 0.0])
    cov = jnp.array([[1.0, -0.3], [-0.3, 1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.isclose(result, 0.0)


def test_x0():
    def f(x):
        return x[0]

    mean = jnp.array([0.0])
    cov = jnp.array([[1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.isclose(result, 0.0)

    mean = jnp.array([1.0, 0.0])
    cov = jnp.array([[1.0, -0.3], [-0.3, 1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.isclose(result, 1.0)


def test_corr():
    def f(x):
        return x[0] * x[1]

    mean = jnp.array([0.0, 0.0])
    cov = jnp.array([[1.0, -0.3], [-0.3, 1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.isclose(result, -0.3)


def test_matrix_integrand():
    def f(x):
        return jnp.outer(x, x)

    mean = jnp.array([0.0, 0.0, 0.0])
    cov = jnp.array([[1.0, -0.3, 0.2], [-0.3, 1.0, 0.1], [0.2, 0.1, 1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.allclose(result, cov)
